# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import json
import shutil
import tarfile
from operator import itemgetter
from pathlib import Path

import pytest
from conftest import (
    ADDITIONAL_METADATA_FILENAME,
    ANCILLARY_PRODUCTS,
    EVLA_EB_LOCATOR,
    EVLA_EB_NAME,
    EVLA_EB_PRODUCT_GROUP,
    IMAGE_LOCATOR,
    IMAGE_PRODUCT_GROUP,
    find_example_manifest,
    find_ingestion_artifacts_tar,
    populate_fake_evla_eb_curator_source_path,
    populate_fake_tmpx_ratuqh_ingest_path,
)
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.manifest_components import (
    CURATOR_MANIFEST_FILENAME,
    INGESTION_ARTIFACTS_NAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    IngestionManifestKey,
    ParamsKey,
    ReingestGroup,
    ReingestKey,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import (
    AncillaryProductType,
    CuratorType,
    IngestionManifestException,
    ScienceProductType,
    Telescope,
)

PRODUCT_SORT_KEY = itemgetter("filename")


@pytest.mark.parametrize("use_alternate_manifest_destination", [True, False])
def test_manifest_full_curation_eb_manifest(
    use_alternate_manifest_destination: bool, ingest_path: Path, alternate_manifest_destination: Path
):
    """Test of manifest creation and serialization for full curation of an execution block

    :param use_alternate_manifest_destination: whether to stick the manifest in the same place as curator's sources
    :param ingest_path: directory where curator's sources will be
    :param alternate_manifest_destination: directory to stick the manifest into if use_alternate_manifest_destination
    """
    eb_files = populate_fake_evla_eb_curator_source_path(ingest_path)

    manifest_destination_dir = ingest_path
    if use_alternate_manifest_destination:
        manifest_destination_dir = alternate_manifest_destination
    assert len(eb_files) == 1
    eb_dir = eb_files[0]
    actual_manifest, actual_manifest_file = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        manifest_source_dir=eb_dir,
        sp_type=ScienceProductType.EXEC_BLOCK.value,
        locator=EVLA_EB_LOCATOR,
        filename=EVLA_EB_NAME,
        curate=(CuratorType.FULL, None),
        manifest_destination_dir=manifest_destination_dir,
        product_group=EVLA_EB_PRODUCT_GROUP,
    ).build()
    assert actual_manifest_file.name == CURATOR_MANIFEST_FILENAME
    assert actual_manifest_file.parent == manifest_destination_dir

    # Check metadata
    params = actual_manifest.parameters
    assert params.ngas_ingest is False
    assert params.telescope == Telescope.EVLA
    assert params.staging_source_dir == eb_dir

    assert actual_manifest.input_group
    assert not actual_manifest.input_group.science_products

    output_group = actual_manifest.output_group
    assert output_group
    assert len(output_group.science_products) == 1
    assert output_group.science_products[0].filename == EVLA_EB_NAME
    assert len(output_group.ancillary_products) == 1
    assert INGESTION_ARTIFACTS_NAME in output_group.ancillary_products[0].filename

    reingest = actual_manifest.reingest
    assert reingest
    assert reingest.targets == None
    assert reingest.locator == None
    assert reingest.product_group == EVLA_EB_PRODUCT_GROUP
    assert reingest.product_type == ScienceProductType.EXEC_BLOCK

    # Check actual file
    with open(actual_manifest_file) as f:
        actual_manifest_deser = json.load(f)
    assert actual_manifest.to_dict() == actual_manifest_deser

    # Check against expected manifest
    expected_manifest_file = find_example_manifest("full_curation_evla_eb_manifest")
    with open(expected_manifest_file) as f:
        expected_manifest = json.load(f)

    # ingestion_path depends on the ingest_path fixture, so ignore it
    expected_params = expected_manifest["parameters"]
    actual_params = actual_manifest_deser["parameters"]
    assert actual_params["ngas_ingest"] == expected_params["ngas_ingest"]
    assert actual_manifest_deser["reingest"] == expected_manifest["reingest"]
    assert actual_manifest_deser.get("input_group") == expected_manifest.get("input_group")

    expected_outgroup = expected_manifest["output_group"]
    expected_osp = expected_outgroup["science_products"]
    actual_outgroup = actual_manifest_deser["output_group"]
    actual_osp = actual_outgroup["science_products"]
    assert actual_osp == expected_osp

    # the ingestion_artifacts' filename depends on current time, so ignore it
    expected_aps = expected_outgroup["ancillary_products"]
    actual_aps = actual_outgroup["ancillary_products"]
    assert len(expected_aps) == len(actual_aps)
    for expected_ap, actual_ap in zip(expected_aps, actual_aps):
        assert expected_ap["type"] == actual_ap["type"]

    # get ingestion artifacts
    artifacts_file = [file for file in manifest_destination_dir.glob("ingestion_artifacts*.tar")][0]
    with tarfile.open(artifacts_file, "r") as tar:
        # confirm that contains has as many files as we expect...
        members = tar.getmembers()

    assert len(members) == 1
    member = members[0]
    mf_path = Path(member.name)
    assert mf_path.name == CURATOR_MANIFEST_FILENAME


@pytest.mark.parametrize("has_curation_source", [True, False])
def test_manifest_partial_curation_eb(has_curation_source: bool, ingest_path: Path):
    """Test full-curation manifest creation & serialization
    :param ingest_path: directory where curator's sources will be
    """
    eb_files = populate_fake_evla_eb_curator_source_path(ingest_path)
    assert len(eb_files) == 1
    eb_dir = eb_files[0]

    curation_source = eb_dir
    expected_manifest_name = "partial_curation_evla_eb_manifest"
    if not has_curation_source:
        expected_manifest_name = "partial_curation_evla_eb_manifest_no_curation_source"
        curation_source = None

    actual_manifest, actual_manifest_file = IngestionManifestBuilder(
        telescope=Telescope.EVLA.value,
        manifest_source_dir=curation_source,
        sp_type=ScienceProductType.EXEC_BLOCK.value,
        locator=EVLA_EB_LOCATOR,
        filename=EVLA_EB_NAME,
        curate=(CuratorType.PARTIAL, ["subscans.dec"]),
        manifest_destination_dir=ingest_path,
    ).build()

    # Check metadata
    params = actual_manifest.parameters
    assert params.ngas_ingest is False
    assert params.telescope == Telescope.EVLA
    if has_curation_source:
        assert params.staging_source_dir == curation_source
    else:
        assert params.staging_source_dir is None
    assert not actual_manifest.input_group
    assert not actual_manifest.output_group
    reingest = actual_manifest.reingest
    assert reingest
    assert reingest.targets is not None
    assert reingest.locator == EVLA_EB_LOCATOR
    assert reingest.product_type == ScienceProductType.EXEC_BLOCK
    assert reingest.product_group == None

    # Check that manifest file exists on disk
    with open(actual_manifest_file) as f:
        actual_manifest_deser = json.load(f)
    assert actual_manifest.to_dict() == actual_manifest_deser

    # Check against expected manifest
    expected_manifest_file = find_example_manifest(expected_manifest_name)
    with open(expected_manifest_file) as f:
        expected_manifest = json.load(f)
    assert actual_manifest_deser.get("input_group") == expected_manifest.get("input_group")
    assert actual_manifest_deser.get("output_group") == expected_manifest.get("output_group")
    assert actual_manifest_deser["reingest"] == expected_manifest["reingest"]

    # Make sure there are no artifacts
    artifacts_file = [file for file in ingest_path.glob("ingestion_artifacts*.tar")]
    assert not artifacts_file


@pytest.mark.parametrize("is_full_curation", [True, False])
def test_curation_manifest_bad_no_manifest_destination_directory(is_full_curation: bool, ingest_path: Path):
    curator_type = CuratorType.PARTIAL
    target_list = ["subscans.ra"]
    if is_full_curation:
        curator_type = CuratorType.FULL
        target_list = None
    populate_fake_evla_eb_curator_source_path(ingest_path)
    with pytest.raises(IngestionManifestException, match="directory to house the manifest"):
        IngestionManifestBuilder(
            telescope=Telescope.EVLA.value,
            sp_type=ScienceProductType.EXEC_BLOCK.value,
            locator=EVLA_EB_LOCATOR,
            filename=EVLA_EB_NAME,
            curate=(curator_type, target_list),
            manifest_source_dir=None,
            manifest_destination_dir=None,
        )


@pytest.mark.parametrize("use_file_list", [True, False])
def test_manifest_full_curation_image(use_file_list: bool, ingest_path: Path, alternate_manifest_destination: Path):
    """Test manifest creation and serialization for full curation of an image
    :param ingest_path: directory where curator's sources will be
    """
    image_paths = populate_fake_tmpx_ratuqh_ingest_path(ingest_path, is_for_curation_test=True)

    curation_source = ingest_path
    file_list = None
    manifest_destination_dir = ingest_path
    if use_file_list:
        shutil.rmtree(ingest_path)
        manifest_destination_dir = alternate_manifest_destination
        file_list = [image_path.name for image_path in image_paths]
        curation_source = None

    actual_manifest, actual_manifest_file = IngestionManifestBuilder(
        file_list=file_list,
        manifest_destination_dir=manifest_destination_dir,
        product_group=IMAGE_PRODUCT_GROUP,
        input_group_locator=IMAGE_LOCATOR,
        manifest_source_dir=curation_source,
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator=IMAGE_LOCATOR,
        curate=(CuratorType.FULL, None),
    ).build()
    assert actual_manifest_file.name == CURATOR_MANIFEST_FILENAME

    # Check metadata
    mf_json = actual_manifest.to_dict()
    keys = ["parameters", "input_group", "output_group", "reingest"]
    assert len(mf_json) == len(keys)
    for key in keys:
        assert key in mf_json.keys()
    params = actual_manifest.parameters
    assert params.ngas_ingest is False
    assert params.telescope == Telescope.EVLA
    if use_file_list:
        assert params.staging_source_dir is None
    else:
        assert params.staging_source_dir == ingest_path
    reingest = actual_manifest.reingest
    assert reingest
    assert reingest.targets == None
    assert reingest.locator == None
    assert reingest.product_group == IMAGE_PRODUCT_GROUP
    assert reingest.product_type == ScienceProductType.IMAGE
    assert actual_manifest.input_group
    assert len(actual_manifest.input_group.science_products) == 1

    """
    The ancillary_products section of the manifest we build
    should match the one in the example.
    """
    output_group = actual_manifest.output_group
    assert output_group
    aps = output_group.ancillary_products
    assert aps
    assert len(aps) == len(ANCILLARY_PRODUCTS)

    # Check contents of ancillary_products
    weblog_candidates = [ap for ap in aps if ap.filename == WEBLOG_FILENAME]
    assert len(weblog_candidates) == 1
    assert weblog_candidates[0].type == AncillaryProductType.PIPELINE_WEBLOG

    ingest_artifacts_tar = find_ingestion_artifacts_tar(manifest_destination_dir)
    ingest_artifacts_candidates = [ap for ap in aps if ap.filename == ingest_artifacts_tar.name]
    assert len(ingest_artifacts_candidates) == 1
    assert ingest_artifacts_candidates[0].type == AncillaryProductType.INGESTION_ARTIFACTS
    """
    The ingestion_artifacts tar should contain ONLY the ingestion manifest
    """
    with tarfile.open(ingest_artifacts_tar, "r") as tar:
        members = tar.getmembers()
        assert len(members) == 1
        for member in members:
            assert member.name.endswith(CURATOR_MANIFEST_FILENAME)

    # Inspect the manifest's JSON dict
    """
    The output_group section of the manifest we build
    should match the one in the example:
    * a "science_products" section containing two science products comprising "type", "filename",
    and six ancillary products on each science product
    * an "ancillary products" section comprising three ancillary products
    """
    mf_json = actual_manifest.to_dict()
    if use_file_list:
        assert ParamsKey.INGESTION_PATH.value not in mf_json[IngestionManifestKey.PARAMETERS.value].keys()
    og_json = mf_json[IngestionManifestKey.OUTPUT_GROUP.value]
    assert len(og_json[IngestionManifestKey.SCIENCE_PRODUCTS.value]) == 2

    for sp_json in og_json[IngestionManifestKey.SCIENCE_PRODUCTS.value]:
        assert len(sp_json) == 3
        assert set(sp_json.keys()) == {IngestionManifestKey.ANCILLARY_PRODUCTS.value, "type", "filename"}
        # and ancillary products belonging to the first science product...
        sp_ap_jsons = sp_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
        assert len(sp_ap_jsons) == 6

    # ... and ancillary products twisting in the wind all by themselves
    ap_jsons = og_json[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
    assert len(ap_jsons) == 3

    # there should be a weblog, the ingestion artifacts tar, and the pipeline artifacts tar
    filenames_found = []
    for ap_json in ap_jsons:
        filename = ap_json["filename"]
        if filename.endswith(TARFILE_EXT):
            if filename.startswith(AncillaryProductType.INGESTION_ARTIFACTS.value) or filename.startswith(
                AncillaryProductType.PIPELINE_ARTIFACTS.value
            ):
                filenames_found.append(filename)
        elif filename == WEBLOG_FILENAME:
            filenames_found.append(filename)

    assert len(filenames_found) == 3

    # Check that manifest file exists on disk
    with open(actual_manifest_file) as f:
        actual_manifest_deser = json.load(f)
    assert actual_manifest.to_dict() == actual_manifest_deser

    # Check against expected manifest
    expected_manifest_file = find_example_manifest("full_curation_evla_image_manifest")
    with open(expected_manifest_file) as f:
        expected_manifest = json.load(f)

    assert actual_manifest_deser.keys() == expected_manifest.keys()
    # ingestion_path depends on the ingest_path fixture, so ignore it
    expected_params = expected_manifest["parameters"]
    actual_params = actual_manifest_deser["parameters"]
    assert actual_params["ngas_ingest"] == expected_params["ngas_ingest"]
    assert expected_manifest["input_group"] == actual_manifest_deser["input_group"]

    expected_outgroup = expected_manifest["output_group"]
    expected_osps = expected_outgroup["science_products"]
    actual_outgroup = actual_manifest_deser["output_group"]
    actual_osps = actual_outgroup["science_products"]
    assert len(actual_osps) == len(expected_osps)
    for actual_osp, expected_osp in zip(
        sorted(actual_osps, key=PRODUCT_SORT_KEY), sorted(expected_osps, key=PRODUCT_SORT_KEY)
    ):
        assert actual_osp["type"] == expected_osp["type"]
        assert actual_osp["filename"] == expected_osp["filename"]
        assert sorted(actual_osp["ancillary_products"], key=PRODUCT_SORT_KEY) == sorted(
            expected_osp["ancillary_products"], key=PRODUCT_SORT_KEY
        )

    # the ingestion_artifacts' filename depends on current time, so ignore it
    expected_aps = expected_outgroup["ancillary_products"]
    actual_aps = actual_outgroup["ancillary_products"]
    assert len(expected_aps) == len(actual_aps)
    for expected_ap, actual_ap in zip(
        sorted(expected_aps, key=PRODUCT_SORT_KEY), sorted(actual_aps, key=PRODUCT_SORT_KEY)
    ):
        assert expected_ap["type"] == actual_ap["type"]
        if expected_ap["type"] != AncillaryProductType.INGESTION_ARTIFACTS.value:
            assert expected_ap["filename"] == actual_ap["filename"]


def test_reingest_block_json_well_formed_partial_curation():
    """
    Make sure our ReingestGroup makes nice JSON without targets and with locator instead of product_group.
    """
    reingest = ReingestGroup(
        product_type=ScienceProductType.EXEC_BLOCK, locator=EVLA_EB_LOCATOR, targets=["subscans.ra", "subscans.dec"]
    )
    reingest_json = reingest.to_dict()
    # Ensure that None fields are absent
    assert ReingestKey.PRODUCT_GROUP.value not in reingest_json.keys()
    # if we can dump it, it's good
    json.dumps(reingest_json)


def test_reingest_block_json_well_formed_full_curation():
    """
    Make sure our ReingestGroup makes nice JSON without targets and with product_group instead of locator.

    Curator requires the "targets" field to be None, not [], in order to trigger full curation.
    """
    reingest = ReingestGroup(product_type=ScienceProductType.EXEC_BLOCK, product_group=EVLA_EB_PRODUCT_GROUP)
    reingest_json = reingest.to_dict()
    # Ensure that None fields are absent
    assert ReingestKey.TARGETS.value not in reingest_json.keys()
    assert ReingestKey.LOCATOR.value not in reingest_json.keys()
    # if we can dump it, it's good
    json.dumps(reingest_json)
